#! /usr/bin/env python
#*******************************************************************************
# ALMA - Atacama Large Millimiter Array
# (c) Associated Universities Inc., 2009 
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
# "@(#) $Id: FocusCalObservation.py 214237 2015-02-06 05:16:28Z nphillip $"

#
# forcing global imports is due to an OSS problem
#
global Control
import Control

global ControlExceptionsImpl
import ControlExceptionsImpl

global Observation
import Observation.ObsCalBase
import Observation.PhaseCalTarget
import Observation.PrimaryBeamCalTarget
global ArraySplit
from Observation.TargetUtility import ArraySplit

class ObsCalAstroHolo(Observation.ObsCalBase.ObsCalBase):

    def __init__(self):
        Observation.ObsCalBase.ObsCalBase.__init__(self)
        self._srcPointFocus = None
        self.refAntennas = []

    def parseOptions(self):
        self.primaryBeamSubscanDuration = self.getExpertParameterOrDefault("SubscanDuration", float, 4.608)
        self.phaseSubscanDuration    = self.getExpertParameterOrDefault("PhaseSubscanDuration", float, 60.0)
        self.dumpDuration            = self.getExpertParameterOrDefault("dumpDuration", float, 0.048)
        self.channelAverageDuration  = self.getExpertParameterOrDefault("channelAverageDuration", float, 0.048)
        self.integrationDuration     = self.getExpertParameterOrDefault("integrationDuration", float, 1.536)
        self.tpIntegrationDuration   = self.getExpertParameterOrDefault("tpIntegrationDuration", float, 0.016)
        self.elLimit                 = self.getExpertParameterOrDefault("ElLimit", str, "20 deg")
        self.pointFocusBand          = self.getExpertParameterOrDefault("pointFocusBand", int, 6)
        self.userSpecifiedSource     = self.getExpertParameterOrDefault("sourceName", str, "Neptune")
        self.numStarSubscans         = self.getExpertParameterOrDefault("numStarSubscans", int, 24)
        self.numStarRepeats          = self.getExpertParameterOrDefault("numStarRepeats", int, 2)
        self.refAntGroup             = self.getExpertParameterOrDefault("refAntGroup", int, 1)
        #self.refAntList              = self.getExpertParameterOrDefault("refAntList", str, "DA57,DV22")
        self.refAntList              = self.getExpertParameterOrDefault("refAntList", str, "DA52,DV18")


    def defineRefAnts(self):
        antennaList = self._array.antennas()
        self.logInfo("antennaList: %s" % str(antennaList))
        if len(self.refAntList) > 0:
             refAntennas = [self.refAntList.split(',')]
        else:
            parts = ArraySplit(antennaList).split()
            self.logInfo("parts: %s" % str(parts))
            if self.refAntGroup == 0:
                refAntennas = parts
            else:
                refAntennas = [parts[self.refAntGroup-1]]
        # Check to make sure the antennas are really in the array
        for ref in refAntennas:
            for antenna in ref:
                if not antenna in antennaList:
                    msg = "Antenna " + antenna + " is not in the array"
                    ex = ControlExceptionsImpl.IllegalParameterErrorExImpl()
                    ex.setData(Control.EX_USER_ERROR_MSG, msg)
                    self.logError(msg)
                    raise ex
        self.logInfo("refAntennas: %s" % str(refAntennas))
        self.refAntennas = refAntennas


    def generateTunings(self):
        corrType = self._array.getCorrelatorType()
        self._pointFocusSpectralSpec = self._tuningHelper.GenerateSpectralSpec(
                band = self.pointFocusBand,
                intent = "interferometry_continuum",
                pol = "4",
                LO_offset_strategy = 2,
                corrType = corrType,
                # ICT-5415 Enabling dual mode causes problems for CLIC
                dualMode = False,
                dump = self.dumpDuration,
                channelAverage = self.channelAverageDuration,
                integration = self.integrationDuration,
                tpSampleTime = self.tpIntegrationDuration)
        self._pointFocusSpectralSpec.name = "Band %d pointing/focus" % self.pointFocusBand


    def doPhaseCal(self):
        try:
            phaseCal = Observation.PhaseCalTarget.PhaseCalTarget(self._srcPointFocus, self._pointFocusSpectralSpec)
            phaseCal.setSubscanDuration(self.phaseSubscanDuration)
            phaseCal.setDataOrigin('CHANNEL_AVERAGE_CROSS')
            phaseCal.setIntegrationTime(1.0)
            phaseCal.setDelayCalReduction(True)
            self.logInfo('Executing PhaseCal on ' + self._srcPointFocus.sourceName + '...')
            phaseCal.execute(self._obsmode)
            self.logInfo('Completed PhaseCal on ' + self._srcPointFocus.sourceName)
        except BaseException, ex:
            print ex
            msg = "Error executing phase on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
                    

    def doPrimaryBeam(self):
        try:
            primaryBeamCal = Observation.PrimaryBeamCalTarget.PrimaryBeamCalTarget(self._srcPointFocus, self._pointFocusSpectralSpec)
            primaryBeamCal.setSubscanDuration(self.primaryBeamSubscanDuration)
            primaryBeamCal.setNumSubscans(self.numStarSubscans)
            primaryBeamCal.setIntegrationTime(self.numStarRepeats * self.numStarSubscans * self.primaryBeamSubscanDuration - 1.0)
            primaryBeamCal.setDataOrigin('CHANNEL_AVERAGE_CROSS')
            for ref in self.refAntennas:
                self.logInfo("Setting Reference antennas to: " + str(ref)) 
                primaryBeamCal.setReferenceAntennas(ref)
                self.logInfo('Executing PrimaryBeamCal on ' + self._srcPointFocus.sourceName + '...')
                primaryBeamCal.execute(self._obsmode)
                self.logInfo('Completed PrimaryBeamCal on ' + self._srcPointFocus.sourceName)
        except BaseException, ex:
            print ex
            msg = "Error executing primaryBeam on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
                    


Observation.ObsCalBase.checkArray()

astroholo = ObsCalAstroHolo()
astroholo.parseOptions()
astroholo.checkAntennas()
astroholo.startPrepareForExecution()
astroholo.generateTunings()
astroholo.findPointFocusSource(minEl=30.0, useThisSource=astroholo.userSpecifiedSource)
astroholo.defineRefAnts()
astroholo.completePrepareForExecution()
astroholo.logInfo("Executing first a short phaseCal for the WVR......")
astroholo.doPhaseCal()
astroholo.logInfo("Executing the primary beam scan(s)......")
astroholo.doPrimaryBeam()
astroholo.closeExecution()

