#! /usr/bin/env python
#*******************************************************************************
# ALMA - Atacama Large Millimiter Array
# (c) Associated Universities Inc., 2009 
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
# "@(#) $Id: ObsCalGrids.py 232822 2016-06-15 02:06:31Z ahirota $"

#
# forcing global imports is due to an OSS problem
#
global sys
import sys
global math
import math
global datetime
import datetime

global CCL
import CCL.Global
import CCL.FieldSource

global Control
import Control

global ControlExceptionsImpl
import ControlExceptionsImpl

global AcsutilPy
import AcsutilPy.FindFile

global Observation
import Observation.AtmCalTarget
import Observation.DelayCalTarget
import Observation.PointingCalTarget
import Observation.FocusCalTarget
import Observation.PhaseCalTarget
import Observation.BandpassCalTarget
import Observation.ObsCalBase
import Observation.ObsCalSources
import Observation.ShadowingChecker


class ObsCalGrids(Observation.ObsCalBase.ObsCalBase):

    azoffset = math.radians(150.0/3600.0)

    def __init__(self):
        Observation.ObsCalBase.ObsCalBase.__init__(self)
        self._srcPointFocus = None
        self._reverseSpecs = False
        self.doneBandpass = False
        self.donePrimaryAmpCal = False
        antennas = self._array.antennas()
        self._shadowingChecker = Observation.ShadowingChecker.ShadowingChecker(useSSRLogger=True,
                                                                               usedAntennaList=antennas)
    def pipelineCheck(self):
        self.logInfo("Checking setup is compatible with PIPELINE requirements...")
        ants = self._array.antennas()
        errStr = ""
        num7m = 0
        num12m = 0
        if len(ants) < 2:
            errStr += " Array has less than two antennas."
        for a in ants:
            if a.startswith("CM"):
               num7m += 1
            else:
               num12m += 1
        if num7m > 0 and num12m > 0:
            errStr += " Array contains both 7m and 12m antennas."
        if len(self.bandList) > 1:
            errStr += " Observing band list contains more than one band (%s)." % str(self.bandList)
        if len(errStr) > 0:
            msg = "Pipeline-friendly mode requires all antnenas to be the same size, and only one observing band. Aborting due to:%s" % errStr
            ex = ControlExceptionsImpl.IllegalParameterErrorExImpl()
            ex.setData(Control.EX_USER_ERROR_MSG, msg)
            self.logError(msg)
            raise ex
        self.logInfo("Setup is compatible with PIPELINE requirements")

    def parseOptions(self):
        self.repeatCount             = self.getExpertParameterOrDefault("RepeatCount", int, 1)
        self.pointingSubscanDuration = self.getExpertParameterOrDefault("PointingSubscanDuration", float, 5.76)
        self.atmSubscanDuration      = self.getExpertParameterOrDefault("AtmSubscanDuration", float, 5.76)
        self.sbrSubscanDuration      = self.getExpertParameterOrDefault("SBRSubscanDuration", float, 5.76)
        self.focusSubscanDuration    = self.getExpertParameterOrDefault("FocusSubscanDuration", float, 5.76)
        self.dumpDuration            = self.getExpertParameterOrDefault("dumpDuration", float, 0.192)
        self.channelAverageDuration  = self.getExpertParameterOrDefault("channelAverageDuration", float, 0.576)
        self.integrationDuration     = self.getExpertParameterOrDefault("integrationDuration", float, 2.88)
        #self.atmIntegrationDuration  = self.getExpertParameterOrDefault("AtmIntegrationDuration", float, 0.576)
        self.tpIntegrationDuration   = self.getExpertParameterOrDefault("tpIntegrationDuration", float, 0.016)
        self.elLimit                 = self.getExpertParameterOrDefault("ElLimit", str, "20 deg")
        self.HourAngleLimit          = self.getExpertParameterOrDefault("HourAngleLimit", float, 3.0)
        self.numTargets              = self.getExpertParameterOrDefault("NumTargets", int, 10)
        self.pointFocusBand          = self.getExpertParameterOrDefault("pointFocusBand", int, 7)
        self.pipelineFriendly        = self.getExpertParameterOrDefault("pipelineFriendly", int, 0)
        bandStr                      = self.getExpertParameterOrDefault("bandList", str, "7,3")
        self.bandList = []
        for s in bandStr.split(','):
            n = int(s)
            if n < 1 or n > 10:
                raise Exception("Invalid band number in band list: '%s'" % s)
            self.bandList.append(n)
        self.logInfo("Band list: %s" % str(self.bandList))
        if self.pipelineFriendly:
            self.pipelineCheck()

    def generateTunings(self):
        corrType = self._array.getCorrelatorType()
        self._pointFocusSpectralSpec = self._tuningHelper.GenerateSpectralSpec(
                band = self.pointFocusBand,
                intent = "interferometry_continuum",
                corrType = corrType,
                dualMode = True,
                dump = self.dumpDuration,
                channelAverage = self.channelAverageDuration,
                integration = self.integrationDuration,
                tpSampleTime = self.tpIntegrationDuration)
        self._pointFocusSpectralSpec.name = "Band %d pointing/focus" % self.pointFocusBand
        self._calSpectralSpecs = []
        for band in self.bandList:
            ss = self._tuningHelper.GenerateSpectralSpec(
                            band = band,
                            intent = "calsurvey",
                            corrType = corrType,
                            dualMode = True,
                            dump = self.dumpDuration,
                            channelAverage = self.channelAverageDuration,
                            integration = self.integrationDuration,
                            tpSampleTime = self.tpIntegrationDuration)
            ss.name = "Band %d calsurvey" % band
            self._calSpectralSpecs.append(ss)

    def pickupAvailableSSOs(self):
        """
        ICT-7107: pipeline friendly way of selecting SSO
        """
        import Observation.CalibratorCatalog
        cc = Observation.CalibratorCatalog.CalibratorCatalog('observing')
        sources = None
        for ss in self._calSpectralSpecs:
            wavelength = 299792458. / ss.getMeanFrequency()
            specSourcs = cc.getAvailableSSOs(wavelength,
                                             minEl=40.,
                                             maxEl=80.,
                                             duration=1200.,
                                             nRequest=-1,
                                             checkSourceNumbers=False,
                                             maxShadowFraction=0.2,
                                             verbose=False)

            srcNameStr = ",".join([source.getName() for source in specSourcs])
            self.logInfo("Picked up %d sources (%s) for spectral spec '%s'" % (len(specSourcs), srcNameStr, ss.name))
            if sources is None:
                sources = set(specSourcs)
            else:
                sources = sources.intersection(specSourcs)
        srcList = [source.buildFieldSource() for source in sources]
        return srcList

    def populateSourceList(self, nameList=None):
        self._calSources = [self._srcPointFocus]
        if not self.pipelineFriendly:
            self._calSources += self.sourceHelper.getUnobscuredPrimaryFluxCalSources()
        else:
            ssoSources = self.pickupAvailableSSOs()
            if len(ssoSources) == 0 :
                raise Exception("Currently cannot find out a primary flux calibrator in a pipeline-friendly way.")
            self._calSources.append(ssoSources[0])
        self._calSources += self.sourceHelper.getAllGridSources()
        for i in range(1, len(self._calSources)):
            if self._calSources[i].sourceName == self._srcPointFocus.sourceName:
                self.logInfo("Deleting duplicate source list entry %s" % self._calSources[i].sourceName)
                del self._calSources[i]
                break
        self.logInfo("Source list:")
        for iSrc, src in enumerate(self._calSources):
            self.logInfo("[%d] %s" % (iSrc, str(src.sourceName)))

    def orderedSpecs(self):
        ret = self._calSpectralSpecs
        if self._reverseSpecs:
            ret = reversed(self._calSpectralSpecs)
        self._reverseSpecs = not self._reverseSpecs
        return ret

    def doPointing(self):
        try:
            pointingCal = Observation.PointingCalTarget.PointingCalTarget(self._srcPointFocus, self._pointFocusSpectralSpec)
            pointingCal.setSubscanDuration(self.pointingSubscanDuration)
            pointingCal.setDataOrigin('CHANNEL_AVERAGE_CROSS')
            if not self.pipelineFriendly:
                pointingCal.setDelayCalReduction(True)
            self.logInfo('Executing PointingCal on ' + self._srcPointFocus.sourceName + '...')
            pointingCal.execute(self._obsmode)
            self.logInfo('Completed PointingCal on ' + self._srcPointFocus.sourceName)
            result = pointingCal.checkResult(self._array)
            self.logInfo("Result is: %s" % str(result))
            if len(result) > 0: 
                for key in result.keys():
                    self.logInfo("Found solution for %s using polarization(s) %s" %
                            (key, result[key]))
                pointingCal.applyResult(self._obsmode, result)
            else:
                if not "OSS" in self._array._arrayName:
                    raise Exception("No pointing results!")
        except BaseException, ex:
            print ex
            msg = "Error executing pointing on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
                    
    def doFocus(self):
        try:
            focusCal = Observation.FocusCalTarget.FocusCalTarget(
                    SubscanFieldSource = self._srcPointFocus,
                    Axis = 'Z',
                    SpectralSpec = self._pointFocusSpectralSpec,
                    DataOrigin = 'CHANNEL_AVERAGE_CROSS',
                    SubscanDuration = self.focusSubscanDuration,
                    OneWay = False,
                    NumPositions = 7)
            self.logInfo('Executing FocusCal on ' + self._srcPointFocus.sourceName + '...')
            focusCal.execute(self._obsmode)
            self.logInfo('Completed FocusCal on ' + self._srcPointFocus.sourceName)
            result = focusCal.checkResult(self._array)
            self.logInfo("Result is: %s" % str(result))
            if len(result) > 0: 
                for key in result.keys():
                    self.logInfo("Found solution for %s using polarization(s) %s" %
                            (key, result[key]))
                focusCal.applyResult(self._obsmode, result)
            else:
                if not "OSS" in self._array._arrayName:
                    raise Exception("No focus results!")
        except BaseException, ex:
            print ex
            msg = "Error executing focus on source %s" % self._srcPointFocus.sourceName
            self.logError(msg)
            self.closeExecution(ex)
            raise ex
                    
    def doCalSource(self, src):
        doneBandpassThisTime = False
        donePrimaryAmpCalThisTime = False
        for ss in self.orderedSpecs():
            try:
                atm = Observation.AtmCalTarget.AtmCalTarget(src, ss, doHotLoad=True)
                atm.setOnlineProcessing(True)
                atm.setDataOrigin('FULL_RESOLUTION_AUTO')
                atm.setDoZero(False)
                atm.setSubscanDuration(self.atmSubscanDuration)
                atm.setIntegrationTime(1.5)
                atm.setWVRCalReduction(True)
                # Applying the results takes a while with lots of
                # antennas, so until we use online WVR, don't bother
                atm.setApplyWVR(False)
                # Currently we need to set a reference source to use an offset.
                atm._referenceSource=atm._source
                try:
                        atm._referenceOffset=CCL.SourceOffset.stroke(self.azoffset,0,0,0,Control.HORIZON)
                        self.logInfo("Using 10.6 and later SourceOffset.stroke for AtmCal reference position")
                except:
                        self.logInfo("Using 10.4 and earlier SourceOffset for AtmCal reference position")
                        atm._referenceOffset=Control.SourceOffset(self.azoffset,0,0,0,Control.HORIZON)
                self.logInfo('Executing AtmCal on ' + src.sourceName + '...')
                atm.execute(self._obsmode)
                self.logInfo('Completed AtmCal on ' + src.sourceName)

                subscanDuration = 0.3e-9 * ss.getMeanFrequency()

                if not self.doneBandpass:
                    ampliCal = Observation.BandpassCalTarget.BandpassCalTarget(src, ss)
                    doneBandpassThisTime = True
                elif not self.donePrimaryAmpCal:
                    ampliCal = Observation.AmplitudeCalTarget.AmplitudeCalTarget(src, ss)
                    donePrimaryAmpCalThisTime = True
                else:
                    ampliCal = Observation.PhaseCalTarget.PhaseCalTarget(src, ss)
                ampliCal.setSubscanDuration(subscanDuration)
                ampliCal.setIntegrationTime(1.0)
                self.logInfo('Executing AmplitudeCal on ' + src.sourceName + '...')
                ampliCal.execute(self._obsmode)
                self.logInfo('Completed AmplitudeCal on ' + src.sourceName)
                firstSS = False
            except BaseException, ex:
                print ex
                msg = "Error executing cal survey scans on source %s" % src.sourceName
                self.logError(msg)
                self.closeExecution(ex)
                raise ex
        if doneBandpassThisTime:
            self.doneBandpass = True
        if donePrimaryAmpCalThisTime:
            self.donePrimaryAmpCal = True

    def doCalObservations(self):
        for src in self._calSources:
            self.logInfo("Source name: %s, object: %s" % (src.sourceName, str(src)))
            if isinstance(src, CCL.FieldSource.EquatorialSource):
                lstSec = self.sourceHelper.getLSTsecondOfDay()
                RASec = (24*60*60 / 360) * math.degrees(src.sourceCoordinates.longitude.get())
                # Get hour angle normalised to range -12,+12h
                HA = (lstSec - RASec) / 3600.0
                if(HA <= -12.0): HA += 24.0
                elif(HA > 12.0): HA -= 24.0
                self.logInfo("LST seconds: %d, source RA seconds: %d, hour angle hours: %.3f" % (lstSec, int(RASec), HA))
                if abs(HA) > self.HourAngleLimit:
                    self.logInfo("Skipping source '%s' as hour angle magnitude over limit of %.3f hours" % (src.sourceName, self.HourAngleLimit))
                    continue
                # Check shadowing
                az, el = Observation.ObsCalSources.ObsCalSources.sourceAzEl(src)
                fBlocked = self._shadowingChecker.isBlocked(math.degrees(az), math.degrees(el), verbose=True, returnFraction=True)
                if fBlocked > 0.25:
                    self.logInfo("Skipping source '%s' as %.1f percent of antennas will suffer from shadowing" % (src.sourceName, fBlocked * 100))
                    continue
            if not self.isObservable(src, 600):
                self.logInfo("Skipping source '%s' as not observable" % src.sourceName)
                continue
            self.doCalSource(src)


Observation.ObsCalBase.checkArray()

survey = ObsCalGrids()
survey.parseOptions()
survey.checkAntennas()
survey.startPrepareForExecution()
survey.generateTunings()
survey.findPointFocusSource()
survey.populateSourceList()
survey.completePrepareForExecution()
survey.logInfo("Executing first pointing...")
survey.doPointing()
survey.logInfo("Executing second pointing -- make sure results are good!...")
survey.doPointing()
survey.logInfo("Executing focus...")
survey.doFocus()
survey.logInfo("Executing third pointing after focus -- make sure results are good!...")
survey.doPointing()
survey.logInfo("Executing Calibration observations...")
survey.doCalObservations()
survey.closeExecution()
