#! /usr/bin/env python
# ALMA - Atacama Large Millimeter Array
# (c) Associated Universities Inc., 2009 - 2014
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that i1t will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

#************************************************************************
#   NAME StandardInterferometry.py
# 
#   SYNOPSIS This script is intended to meet the needs of single or multi-target
#       observations which have a common spectral configuration for all
#       science targets and a shared primary phase calibrator.  
# 
#------------------------------------------------------------------------

#
# forcing global imports is due to an OSS problem
#
global CCL
from CCL.Global import getArray
import CCL.APDMSchedBlock
from CCL.ObservingModeBase import SBOutOfTime
from PyDataModelEnumeration import PyCalibrationDevice
global Observation
import Observation.SchedulingBlock
from Observation.SBExecutionMode import SBExecutionMode, SBExecutionError
from Observation.Global import simulatedArray, addClassToMode
from Observation.PointingCalTarget import PointingCalTarget
from Observation.ScanList import ScanList
from Observation.SBExecState import SBExecState
import Observation.SSRLogger


# Ugly hack to get nice logging when we're not written as a class
class StandardInterferometry(Observation.SSRLogger.SSRLogger):
    def __init__(self):
        Observation.SSRLogger.SSRLogger.__init__(self, self.__class__.__name__)
logger = StandardInterferometry()

sb = Observation.SchedulingBlock.SchedulingBlock()
logger.logInfo('Running %s scheduling block' % sb._controlSB.name)

# Get expert parameters as needed here
elLimit         = float(sb.getExpertParameter('ElLimit', default=20.))
maxPointingSeparation = float(sb.getExpertParameter('MaxPointingSeparation', default=25.))
sourceCycle     = float(sb.getExpertParameter('SourceCycleTime', default=0))
useScanSequence = bool(int(sb.getExpertParameter('useScanSequence', default=True)))
useDualMode     = bool(int(sb.getExpertParameter('useDualMode', default=True)))
fastSWEnabled   = bool(int(sb.getExpertParameter('FastSWEnabled', default=False)))
# Below are parameters for DiffGainCalTarget
dgcExtCycle   = float(sb.getExpertParameter('dgcExtCycle', default=7200.))
dgcIntCycle   = float(sb.getExpertParameter('dgcIntCycle', default=60.))
dgcRefIntTime = float(sb.getExpertParameter('dgcRefIntTime', default=120.))
dgcSciIntTime = float(sb.getExpertParameter('dgcSciIntTime', default=180.))
dgcSourceName = str(sb.getExpertParameter('dgcSourceName', default=''))
dgcSourceRa   = str(sb.getExpertParameter('dgcSourceRa', default=''))  # hh:mm:ss.sss
dgcSourceDec  = str(sb.getExpertParameter('dgcSourceDec', default='')) # dd.mm.ss.sss
dgcPostpone   = float(sb.getExpertParameter('dgcPostpone', default=0.))
useDGC        = int(sb.getExpertParameter('useDGC', default=0))
# A parameter for overrding WVR correction on/off/both. ("", "off", "on", "both")
wvrProduct = str(sb.getExpertParameter("wvrProduct", default=''))
# logger.logInfo('%s' % sb.getExpertParameters())

logger.logInfo("useDualMode value %s" % useDualMode)
logger.logInfo("useScanSequence value %s" % useScanSequence)
logger.logInfo('fastSWEnabled=%s' % fastSWEnabled)
logger.logInfo('wvrProduct=%s' % wvrProduct)

B_max = 450.0 # length, in meters, of the longest baseline in the array to be executed.  Could either
              # get this from the resolution requirement or from the scheduler with the configuration
if sourceCycle <= 0:
    sourceCycle = 12.0/(B_max*7.29e-5)
logger.logInfo('Source Cycle Time: %f s' % sourceCycle)
sessionControl = int(sb.getExpertParameter('SessionControl', default=0))
logger.logInfo('Session control %d' % sessionControl)
logger.logInfo('Simulated Array %s' % simulatedArray())

# Get reference to the array and start the exec block
array = getArray()

# In bands 8 and lower it makes most practical sense to point in-band if we
# have weather for science in that band. The practical difficulties of
# maintaining good TMCDB band offsets trump the S/N arguments, which are not
# that great anyway as people forget the positive effect of decreasing beam
# width with increasing frequency.
Observation.SSRTuning.overridePointingSpectralSpecs(sb, array, logger)

array.beginExecution()
obsmode = array.getInterferometryObservingMode()

# Add the SB execution methods
addClassToMode(obsmode, SBExecutionMode)

# Session memory
sbe = SBExecState(array, sb)
if sessionControl:
    sbe.startSession()
obsmode.addSBExecState(sbe)

# Just start on the sky, to be sure. Should be in the mode defaults.
obsmode.setCalibrationDevice(PyCalibrationDevice.NONE)

# Set observing mode restrictions
logger.logInfo("Setting elevation limit to %f degrees" % elLimit)
obsmode.setElevationLimit('%f deg' % elLimit)

endTime = sb.getMaximumExecutionTime()
logger.logInfo("Setting maximum execution time to %s seconds" % (endTime))
obsmode.setMaximumSchedBlockTime(endTime)

logger.logInfo("Setting max. pointing separation  to %f degrees" % maxPointingSeparation)
PointingCalTarget.setMaximumSeparation('%f deg' % maxPointingSeparation )

groupList = sb.getGroupList()
# Check the visibility of science targets in all groups.
for iGroup, group in enumerate(groupList):
    if obsmode.checkScienceTargetVisibility(group):
        continue
    raise SBExecutionError("No visible science target [Group%d]" % (iGroup + 1))

for iGroup, group in enumerate(groupList):
    # Perform doppler correction for spectral specs associated with this group.
    sb.adjustDopplerByGroup(obsmode, group)
    # Then, resolve queries.
    group.prepareTargets()
    sb.updateTargetList()
    # And, create associated ATM targets (all spectral specs should be in TOPO frame)
    group.createAtmCalTargets(isSingleDish=False)
    # SBR targets are omitted from normal science SB executions, but
    # if it is requested to revert, here is the place to add SBR targets.
    ### group.createSBRatioCalTargets()
    sb.updateTargetList()

# Just for in case...
for target in sb.getTargetList():
    dopRef = str(target.getSpectralSpec().FrequencySetup.dopplerReference)
    logger.logInfo("dopRef=%s getUseReferencePosition=%s %s" % (dopRef, target.getUseReferencePosition(), target))
    import Observation.AtmCalTarget
    if isinstance(target, Observation.AtmCalTarget.AtmCalTarget):
        assert(target.getUseReferencePosition() == True)
    else:
        target.setUseReferencePosition(False)

if wvrProduct != "":
    # Switch between online-WVR on/off/both.
    sb.setWVRProduct(wvrProduct)

sb.setTargetsId()

# If "AddToCleanup" target-level expert parameter is defined, add the target to cleanup list.
for group in groupList:
    group.addTargetsToCleanUpIfRequired(obsmode)

if sessionControl:
    # Recycle Field Sources
    obsmode.sbe.recycleFieldSources()
    # Recycle attenuator settings
    # Mark _ifSkyAttenSet to re-use previous attenuator optimizations.
    # This relies on the fact that attenuator setting name is identical as spectral spec name.
    obsmode.sbe.recycleAttenSettings()
    obsmode.sbe.restoreScienceTargetParameters()

# SCIREQ-290: remove ampcal, if it shares same source with bandpass cal
for group in groupList:
    group.removeSharableTargets()

# None means OK. If an error occurs the status is replaced, in the catch block, by the exception.
status = None

try:
    # The first group is for SB calibrators, we don't expect any science target.
    group = groupList[0]
    groupNumber = group.groupIndex
    logger.logInfo("Start Observing Group %d (Calibrations)" % groupNumber)
    scanList = ScanList() if useScanSequence else None
    # Optionally force-enable dual mode on SpectralSpecs.
    if useDualMode :
        for target in sb.getTargetList() :
            ss = target._spectralSpec
            logger.logInfo("Dual mode addition: evaluating SpectralSpec named '%s'" % str(ss.name))
            if not hasattr(ss,'SquareLawSetup') or ss.SquareLawSetup is None:
                logger.logInfo("Dual mode addition: adding SquareLawSetup to SpectralSpec named '%s'" % str(ss.name))
                ss.SquareLawSetup = CCL.APDMSchedBlock.SquareLawSetup()
                ss.SquareLawSetup.integrationDuration.set(0.016)
    # ICT-5552: For Cycle-3 OT will still give us short subscans, but we'll combine them when we can
    for target in sb.getTargetList() :
        logger.logInfo("setting target.coalesceSubscans = True on '%s' it was %s" % (str(target), str(target.coalesceSubscans)))
        target.coalesceSubscans = True

    nFocus, nFocusCleanup = obsmode.executeCalTargetList(group.getFocusCalTargetList())

    # Polarization section needs to be cleanup a bit.
    numDone, numInCleanup = obsmode.executeCalTargetList(group.getPolarizationCalTargetList(),
                                                         deferToCleanupList=True, scanList=scanList)
    if useScanSequence:
        scanList.execute(obsmode)
    if len(group.getPolarizationCalTargetList()) > 0:
        obsmode.addTargetToCleanupList(group.getPolarizationCalTargetList()[0])

    numDone, numInCleanup = obsmode.executeCalTargetListSpec(group.getBandpassCalTargetList(),
                                                             deferToCleanupList=True, scanList=scanList)
    if useScanSequence:
       scanList.execute(obsmode)

    obsmode.executeCalTargetList(group.getDelayCalTargetList(), scanList=scanList)
    if useScanSequence:
        scanList.execute(obsmode)

    numDone, numInCleanup = obsmode.executeCalTargetList(group.getAmplitudeCalTargetList(),
                                                         deferToCleanupList=True, scanList=scanList)
    if useScanSequence:
        scanList.execute(obsmode)

    # Now the rest of the groups should contain science targets.
    for group in groupList[1:]:
        # TODO: For Cycle5?: allow self-calibration
        groupNumber = group.groupIndex
        # The primary phase calibrator is assumed to be the one with the
        # shortest cycle time in the group.  This effectivly sets the scan
        # duration.
        primaryPhaseCal = obsmode.getPrimaryPhaseCal(group)
        if primaryPhaseCal is None:
            logger.logError("No phase calibrator available in Observing Group %s" % groupNumber)
            raise SBExecutionError("There is no available phase calibrator, at this moment.")
        else:
            logger.logInfo("Selected %s as Primary phase calibrator" % primaryPhaseCal)
    
        # ICT-7246: If phase cycle time is less than 150 seconds, enable fast SW mode.
        fastSWEnabled = group.isFastSWRequired()
        logger.logInfo("[Group%d] fastSWEnabled=%s" % (groupNumber, fastSWEnabled))
        logger.logInfo("Start Observing Group %d (Science)" % groupNumber)

        # construct DiffGainCalTarget, if needed
        if useDGC:
            kwargs = dict(
                dgcExtCycle=dgcExtCycle,
                dgcIntCycle=dgcIntCycle,
                dgcRefIntTime=dgcRefIntTime,
                dgcSciIntTime=dgcSciIntTime,
                dgcSourceName=dgcSourceName,
                dgcSourceRa=dgcSourceRa,
                dgcSourceDec=dgcSourceDec,
                dgcPostpone=dgcPostpone,
            )
            isB2B = group.isBandToBand()
            logger.logInfo('isB2B = %s' % (isB2B))
            # Both B2B and BW switching observations can construct DGC in a same way.
            diffGainCalTarget = group.constructDiffGainCalTargetForB2B(obsmode, **kwargs)

        # A group is defined to be complete when all science targets within
        # the group are complete (or unobservable)
        obsmode.assignSubcycleTime(group.getScienceTargetList(), sourceCycle)
        while not group.isComplete(obsmode):
            outerCycleTargets = list(group.getFocusCalTargetList()) + \
                                list(group.getAmplitudeCalTargetList()) + \
                                list(group.getPolarizationCalTargetList()) + \
                                list(group.getBandpassCalTargetList()) + \
                                list(group.getDelayCalTargetList()) + \
                                list(group.getCheckSourceCalTargetList())
            if useDGC:
                outerCycleTargets.append(diffGainCalTarget)

            secondaryPhaseCals = list(group.getPhaseCalTargetList())
            secondaryPhaseCals.remove(primaryPhaseCal)

            # If one of these target are executed, it would be nice
            # to call primary.overwriteNextExecutionTime(value=0.),
            # since in that case, we can make that duration of
            # each science target execution more uniform.

            obsmode.executeCalTargetList(group.getFocusCalTargetList(), scanList=scanList)
            if useScanSequence:
                scanList.execute(obsmode)

            obsmode.executeCalTargetList(group.getPolarizationCalTargetList(), scanList=scanList)
            if useScanSequence:
                scanList.execute(obsmode)

            obsmode.executeCalTargetList(group.getBandpassCalTargetList(), scanList=scanList)
            if useScanSequence:
                scanList.execute(obsmode)

            obsmode.executeCalTargetList(group.getAmplitudeCalTargetList(), scanList=scanList)
            if useScanSequence:
                scanList.execute(obsmode)

            if useDGC:
                # DiffGainCal
                # For B2B observations, it is desired to execute DGC just before the primary phase cal execution.
                isExecuted = diffGainCalTarget.executeIfNeeded(obsmode, scanList=scanList)
                if isExecuted:
                    # Force execution of next primary phase calibrator.
                    primaryPhaseCal.overwriteNextExecutionTime(value=0.)

            obsmode.executeCalTargetList(group.getPhaseCalTargetList(), deferToCleanupList=True,
                                         minimumInCleanupList=len(group.getPhaseCalTargetList()),
                                         scanList=scanList)
            if not fastSWEnabled and useScanSequence:
                scanList.execute(obsmode)

            # Test Source execution.
            # It is executed at every outer loop execution on the target.
            obsmode.executeCalTargetList(group.getDelayCalTargetList() +
                                         group.getCheckSourceCalTargetList(), scanList=scanList)
            if not fastSWEnabled and useScanSequence:
                scanList.execute(obsmode)

            while not group.isComplete(obsmode):
                # Inner-cycle: if scan-sequence is enabled, try to sequence as many as possible science and phase scans.
                # execute science targets until either a calibration is needed
                obsmode.executeScienceTargetList(group, outerCycleTargets + [primaryPhaseCal], scanList=scanList)
                if not fastSWEnabled and useScanSequence:
                    scanList.execute(obsmode)
                # then, phase-cal
                primaryPhaseCal.execute(obsmode, scanList=scanList)

                if not fastSWEnabled and useScanSequence:
                    scanList.execute(obsmode)

                # Check whether the primary phase calibrator will be above the
                # elevation limit for the next cycle.
                # ICT-6300: If cycle time is extremely long, cap the duration
                # for visibility check by the remaining science target
                # integration time.
                duration = min(primaryPhaseCal.getCycleTime(),
                               group.getRemainingIntegrationTime() + primaryPhaseCal.getIntegrationTime())
                if not primaryPhaseCal.isObservable(obsmode, duration=duration, scanList=scanList):
                    primaryPhaseCal = obsmode.getPrimaryPhaseCal(group, primaryPhaseCal)
                    if primaryPhaseCal is None:
                        raise Exception('There is no available phase calibrator in group%d at this moment, and thus stopping this SB execution. If SB does contain at least one valid phase calibrator in group%d, then it is likely that all of them are (judged to be) below the elevation limit.' % (groupNumber, groupNumber))

                if not fastSWEnabled:
                    # If fast-sw disabled, do not try to sequence phase and science targets
                    break
                if useScanSequence and scanList.getCurrentCompletionTime() > 600:
                    break

                # loop-end condition. (If outer-cycle targets need execution, get out of inner loop)
                currentElapsedTime = obsmode.sbe.getElapsedTime() if sessionControl else obsmode.getElapsedTime()
                if useScanSequence:
                    currentElapsedTime += scanList.getCurrentCompletionTime()
                nextReqTimes = [target.getNextRequiredTime() for target in outerCycleTargets + secondaryPhaseCals]
                if len(nextReqTimes) > 0 and min(nextReqTimes) <= currentElapsedTime:
                    break

            # Exexcute sequence of science and primary phase cal scans
            if useScanSequence:
                scanList.execute(obsmode)

        # Group is complete: remove the primaryPhaseCal from cleanup:
        logger.logInfo("Observing Group %s is completed."% groupNumber)

        obsmode.removeTargetFromCleanupList(primaryPhaseCal)

    # All groups are complete: execute the cleanup (e.g. bandpass or amplitude cals)
    logger.logInfo("All groups are completed.")
    logger.logInfo("Executing cleanup list.")
    obsmode.executeCleanupList()
except SBOutOfTime:
    logger.logWarning("Running short of time, forcing cleanup!")
    groupNumber = group.groupIndex
    if sessionControl and groupNumber > 0 and hasattr(group, "_subcycleIndex"):
        currentIndex = sbe.getScienceTargetIndex(groupNumber)
        if currentIndex == group._subcycleIndex:
            group._subcycleIndex = (group._subcycleIndex + 1) % len(group.getScienceTargetList())
        sbe.saveScienceTargetIndex(groupNumber, group._subcycleIndex)
    if useScanSequence:
        scanList.execute(obsmode)
    logger.logInfo("Executing cleanup list.")
    obsmode._maximumTime = None
    obsmode.executeCleanupList()

except Exception, ex:
    logger.logError("Exception caught: %s" % ex)
    status = ex
    raise

finally:
    if sessionControl:
        sbe.endSBExecution()
    try:
        obsmode.resetLimits()
    except:
        logger.logError("Could not reset limits")

    array.endExecution(status)
#
# ___oOo___
